/****************************************************************************
*
*	  Module Filename:  Realised_Cycle_Type_Manager.cpp
*	  Module Name:      Realised_Cycle_Type_Manager
*	  Description:      Represents type of Cycle that has occurred in the Reactor
*                      
*	  Author:           Chris Gordon-Smith
*
*     Copyright (C) 2006, 2008 Chris Gordon-Smith
*     Email: See http://www.simsoup.info/
*
*     This program is free software; you can redistribute it and/or modify it
*     under the terms of the GNU General Public License as published by the
*     Free Software Foundation; either version 2 of the License, or (at your
*     option) any later version.
*
*     This program is distributed in the hope that it will be useful, but
*     WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
*     or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
*     for more details.
*
*     You should have received a copy of the GNU General Public License along
*     with this program; if not, write to the Free Software Foundation, Inc.,
*     59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*
*     See the file Licence.txt for a copy of the GNU General Public License.
*	  Amendment History:-
*
*         Version    Date                 Update Description
*         =======    ===========          ==================
*
*
*
*
****************************************************************************/

//  Comment out to disallow tracing this file
//#define  ENABLE_TRACE_FILE  

#include    "globldef.h"
#include    "Library_Headers.h"

namespace DN {
#include "list.hpp"        // Special purpose - Non STL
}

//-------------- Tracker  Subsystem  -----------------------------------------------------------------------------

#include  "Realised_Cycle_Type_Manager.h"
#include  "Cycle_Type.h"

//--------------  Reactor  Subsystem  ---------------------------------------------------------------------------
#include  "Reactor_Time_API.h"
//---------------------------------------------------------------------------------------------------------------

Realised_Cycle_Type_Manager*  Realised_Cycle_Type_Manager::Instance_Pt = 0;
//================================================================================================================


void  Realised_Cycle_Type_Manager::Clear()

//  Clear the Realised Cycle Type Manager

{
    Last_Stats_Collection_Time = 0;
    Realised_Cycle_Types.clear();
}
//----------------------------------------------------------------------------------------------------------------
void  Realised_Cycle_Type_Manager::Get_Active_Cycle_Type_Stats(int Reactor_Time,
                                                               Active_Cycle_Summary& Stats)

//  Return statistics about Active Cycles in the Active_Cycle_Summary struct

{
    int Cycle_Length_Accumulator = 0;
    int Num_Active_Cycle_Types = 0;
    int Min_Length = INT_MAX;
    int Max_Length = 0;
    int Total_Active_Cycle_Occurrences = 0;
    int Total_Active_Cycle_Interactions = 0;
    int Occurrence_Counter = 0;

    map<string,Realised_Cycle_Type_Params>::iterator  Iter;
    for (Iter = Realised_Cycle_Types.begin();
         Iter not_eq Realised_Cycle_Types.end();
         Iter++)
    BEGIN
        if(Iter->second.Last_Occurrence_Time > Last_Stats_Collection_Time)
        THEN
            Num_Active_Cycle_Types++;   
            int Length = Iter->second.Cycle_Length;
            Cycle_Length_Accumulator += Length;

            Max_Length = max(Length, Max_Length);
            Min_Length = min(Length, Min_Length);

            Occurrence_Counter = Iter->second.Period_Occurrence_Counter;
            Total_Active_Cycle_Occurrences += Occurrence_Counter;
            Total_Active_Cycle_Interactions += (Length * Occurrence_Counter);

//          Reset the Period Occurrence Counter to zero (we are starting a new period)

            Realised_Cycle_Type_Params  Params;
            Params.Last_Occurrence_Time = Iter->second.Last_Occurrence_Time;
            Params.Cycle_Length = Length;
            Params.Period_Occurrence_Counter = 0;
            Realised_Cycle_Types[Iter->first] = Params;

        ENDIF
    END

    Stats.Num_Cycle_Types = Num_Active_Cycle_Types;

    if (Num_Active_Cycle_Types not_eq 0)
    THEN
        Stats.Avg_Length = double(Cycle_Length_Accumulator) / Num_Active_Cycle_Types;
        int Interval = Reactor_Time - Last_Stats_Collection_Time;
        Assert (Interval > 0, "Realised_Cycle_Type_Manager::Get_Active_Cycle_Type_Stats - Non Positive Statistics Inrterval");

        Stats.Avg_Cycle_Rate = double(Total_Active_Cycle_Occurrences) / Interval;
        Stats.Avg_Interaction_Rate = double(Total_Active_Cycle_Interactions) / Interval;
        Stats.Max_Length = Max_Length;
        Stats.Min_Length = Min_Length;

        Assert ((Stats.Avg_Length >= 2) and (Stats.Min_Length >= 2) and (Stats.Max_Length >= 2), 
                "Realised_Cycle_Type_Manager::Get_Active_Cycle_Type_Summary - Invalid Active_Cycle_Type Lengths");
    ELSE
        Stats.Avg_Length = 0.0;
        Stats.Avg_Cycle_Rate = 0.0;
        Stats.Avg_Interaction_Rate = 0.0;
        Stats.Max_Length = 0;
        Stats.Min_Length = 0;
    ENDIF

    Last_Stats_Collection_Time = Reactor_Time;
}
//--------------------------------------------------------------------------------------------------------------

int   Realised_Cycle_Type_Manager::Get_Num_Realised_Cycle_Types()

//  Return the number of Cycle_Types that have been realised.

{
    return Realised_Cycle_Types.size();
}
//---------------------------------------------------------------------------------------------------------------

Realised_Cycle_Type_Manager*  Realised_Cycle_Type_Manager::Get_Pt()

{
    TRACE("Realised_Cycle_Type_Manager::Get_Pt");
    Assert(Instance_Pt != 0, "Realised_Cycle_Type_Manager::Get_Pt - Instance Pointer not yet initialised");
    return Instance_Pt;
}
//---------------------------------------------------------------------------------------------------------------

void Realised_Cycle_Type_Manager::Notify_Cycle_Occurrence(Tracker*  Tracker_Ptr, int Closing_MolType_Id)

{
//    int Time = Reactor::Get_Reactor__Ptr()->Get_Reactor_Time();
    int Time = Reactor_Time_API::Get_Reactor_Time();
    string Cycle_Type_Identifier;
    Realised_Cycle_Type_Params  Params;

    int  Cycle_Path_Length;
    Cycle_Type::Identify_Cycle_Type_And_Length(Tracker_Ptr, Closing_MolType_Id, Cycle_Type_Identifier, Cycle_Path_Length);

    Params.Last_Occurrence_Time = Time;

    Params.Cycle_Length = Cycle_Path_Length;
    map<string,Realised_Cycle_Type_Params>::iterator Iter; 
    Iter = Realised_Cycle_Types.find(Cycle_Type_Identifier);

    if (Iter == Realised_Cycle_Types.end())
    THEN
        Params.Period_Occurrence_Counter = 1;

//      Create an entry in the map. Fatal error if it can't be created.

        if (Realised_Cycle_Types.insert(make_pair(Cycle_Type_Identifier,Params)).second == false)
        THEN
            Assert(true == false, "Realised_Cycle_Type_Manager::Notify_Cycle_Occurrence - Failed to insert");
        ENDIF
    ELSE
        Params.Period_Occurrence_Counter = (*Iter).second.Period_Occurrence_Counter + 1;
        Realised_Cycle_Types[Cycle_Type_Identifier] = Params;
    ENDIF
}
//---------------------------------------------------------------------------------------------------------------

Realised_Cycle_Type_Manager::Realised_Cycle_Type_Manager()

{
    Last_Stats_Collection_Time = 0;
    Instance_Pt = this;
}



