/****************************************************************************
*
*	  Module Filename:  Input_Record.cpp
*	  Module Name:      Input_Record
*	  Description:      Input Record handler: Holds data from a record of input
*                       read from a file, and makes it available, with 
*                       with validations and conversions, throug an API
*                      
*	  Author:           Chris Gordon-Smith
*
*     Copyright (C) 2008, 2008 Chris Gordon-Smith
*     Email: See http://www.simsoup.info/
*
*     This program is free software; you can redistribute it and/or modify it
*     under the terms of the GNU General Public License as published by the
*     Free Software Foundation; either version 2 of the License, or (at your
*     option) any later version.
*
*     This program is distributed in the hope that it will be useful, but
*     WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
*     or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
*     for more details.
*
*     You should have received a copy of the GNU General Public License along
*     with this program; if not, write to the Free Software Foundation, Inc.,
*     59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*
*     See the file Licence.txt for a copy of the GNU General Public License.
*	  Amendment History:-
*
*         Version    Date                 Update Description
*         =======    ===========          ==================
*
*
*
*
****************************************************************************/
//  Comment out to disallow tracing this file
//#define  ENABLE_TRACE_FILE  

#include  "globldef.h"
#include  "Library_Headers.h"

//--------------  Persistent_Data_Manager  Subsystem  ----------------------------------------------------------
#include  "Input_Record.h"
#include  "Simulation_Input.h"
//--------------------------------------------------------------------------------------------------------------

bool Input_Record::Check_All_Attributes_Read(string&  Error_Text)

//  Check that all of the Input Record's attributes have been read.
//  If not, return false and set the first unread attribute as the Error text

{
    TRACE("Input_Record::Check_All_Attributes_Read");
    map <string, list<string> >::iterator  Iter;
    for (Iter = Attribs_And_Data.begin();
         Iter not_eq Attribs_And_Data.end();
         Iter++)
    BEGIN
        bool Attrib_Read_OK = false;
        string Input_Attrib = (*Iter).first;
        for (list <string>::iterator String_Iter = Attribs_Read.begin();
             String_Iter not_eq Attribs_Read.end();
             String_Iter++)
        BEGIN
            if (*String_Iter == Input_Attrib)
            THEN
                Attrib_Read_OK = true;
                break;
            ENDIF
        END
        if (not Attrib_Read_OK)
        THEN
            Error_Text = Input_Attrib;
            return false;
        ENDIF     
    END
    return true;
}
//----------------------------------------------------------------------------------------------------------------

bool    Input_Record::Check_No_Attrib_Strings_Left(string Attribute_Name,  
                                                 list<string>& Attrib_Strings,     
                                                 string& Error_Text)

//  Check that there are no strings remaining in Attrib_Strings
//  Return false and set the Error_Text if there are

{
    TRACE("Input_Record::Check_No_Attrib_Strings_Left");
    if (not Attrib_Strings.empty())
    {
        Error_Text = Error_Text 
                    + "Too many values for "  
                    + String_In_Quotes(Attribute_Name)
                    + " attribute";
        return false;
    }
    return true;
}
//----------------------------------------------------------------------------------------------------------------

bool Input_Record::Get_Attribute_Strings(string AttrName, list<string>& Attrib_Strings, string&  Error_Text)

//  Get the input text string for the specified attribute type. 
//  Return false if the named attribute was not input

{
    TRACE("Input_Record::Get_Attribute_Strings");
    map <string, list<string> >::iterator  Iter;
    Iter = Attribs_And_Data.find(AttrName);
    if (Iter == Attribs_And_Data.end())
    THEN
        Error_Text = "No " + String_In_Quotes(AttrName) + " attribute";
        return false;
    ELSE
        Attrib_Strings = Iter->second;
        Attribs_Read.push_back(AttrName);  //  Remember that we have read this attribute
        TRACEVAL(AttrName);
    ENDIF
    return  true;
}
//----------------------------------------------------------------------------------------------------------------

string Input_Record::Get_Record_As_String()

{
    TRACE("Input_Record::Get_Record_As_String");

    string Record_As_String;
    for (list<string>::iterator  Iter = Input_Record_For_Print.begin();
        Iter not_eq Input_Record_For_Print.end();
        Iter++)
    BEGIN
        Record_As_String += string ("  ") + (*Iter);  
    END
    Record_As_String += string(".  ");
    return Record_As_String;
}
//----------------------------------------------------------------------------------------------------------------
bool Input_Record::Get_Attr_As_NonNegInt(string  AttrName, 
                                         int& Value, 
                                         string&  Error_Text)

//  Get the attribute value as a single non negative integer
//  It is an error if there are multiple attribute strings

{
    TRACE("Input_Record::Get_Attrib_Value_As_Single_NonNegInt");
    list<string>  Attrib_Strings; 
    if (not Get_Attribute_Strings(AttrName, Attrib_Strings, Error_Text)) return false;
    string  String1 = Attrib_Strings.front();;
    Attrib_Strings.pop_front();
    if (not Check_No_Attrib_Strings_Left(AttrName, 
                                         Attrib_Strings, 
                                         Error_Text)) return false;
    if ((not Text_To_Int(String1, Value)) or  Value < 0)
    THEN
        Error_Text = "Invalid " + AttrName + string(" ") 
                     + String_In_Quotes(String1) 
                     + ": Expected a non-negative integer";
        return false;
    ENDIF
    return true;
}
//----------------------------------------------------------------------------------------------------------------

bool    Input_Record::Get_Attr_As_PosInt(string  AttrName, int& Value, string&  Error_Text)

//  Get the attribute value as a single positive integer
//  It is an error if there are multiple attribute strings

{
    TRACE("Input_Record::Get_Attrib_Value_As_Single_PosInt");

    list<string>  Attrib_Strings; 
    if (not Get_Attribute_Strings(AttrName, Attrib_Strings, Error_Text)) return false;
    string  String1 = Attrib_Strings.front();;
    Attrib_Strings.pop_front();
    if (not Check_No_Attrib_Strings_Left(AttrName, 
                                         Attrib_Strings, 
                                         Error_Text)) return false;
    if ((not Text_To_Int(String1, Value)) or  Value < 1)
    THEN
        Error_Text = "Invalid " + AttrName + string(" ") 
                     + String_In_Quotes(String1) 
                     + ": Expected a positive integer";
        return false;
    ENDIF
    return true;
}
//----------------------------------------------------------------------------------------------------------------


bool Input_Record::Get_Attr_As_Rand_Or_Float(string  AttrName,
                                        bool& Rand_Flag,
                                        double&  Value,
                                        string&  Error_Text)

{
    TRACE("Input_Record::Get_Attr_As_Rand_Or_Float");

    list<string>  Attrib_Strings; 
    if (not Get_Attribute_Strings(AttrName, Attrib_Strings, Error_Text)) return false;
    string  String1 = Attrib_Strings.front();;
    Attrib_Strings.pop_front();
    if (not Check_No_Attrib_Strings_Left(AttrName, 
                                         Attrib_Strings, 
                                         Error_Text)) return false;


    if (String1 == "Random")
    THEN
        Rand_Flag = true;
    ELSE
        if ((not Text_To_Float(String1, Value)))
        THEN
            Error_Text = "Invalid " 
                        + AttrName + " " 
                        + String_In_Quotes(String1) 
                        + ": Expected a floating point number (with no leading zeroes), or the text \"Random\"";
            return false;
        ENDIF
        Rand_Flag = false;
    ENDIF
    return true;
}
//----------------------------------------------------------------------------------------------------------------



bool Input_Record::Get_Attr_As_Rand_Or_NonNegInt(string  AttrName, 
                                                 bool& Rand_Flag,
                                                 int& Value, 
                                                 string&  Error_Text)

{
    TRACE("Input_Record::Get_Attr_As_Rand_Or_NonNegInt");

    list<string>  Attrib_Strings; 
    if (not Get_Attribute_Strings(AttrName, Attrib_Strings, Error_Text)) return false;
    string  String1 = Attrib_Strings.front();;
    Attrib_Strings.pop_front();
    if (not Check_No_Attrib_Strings_Left(AttrName, 
                                         Attrib_Strings, 
                                         Error_Text)) return false;

    if (String1 == "Random")
    THEN
        Rand_Flag = true;
    ELSE
        if ((not Text_To_Int(String1, Value)) or  Value < 0)
        THEN
            Error_Text = "Invalid " 
                        + AttrName 
                        + " " 
                        + String_In_Quotes(String1)
                        + ": Expected a non-negative integer, or the text \"Random\"";
            return false;
        ENDIF
        Rand_Flag = false;
    ENDIF
    return true;
}
//----------------------------------------------------------------------------------------------------------------

bool Input_Record::Get_Attr_As_Rand_Or_PosInt(string  AttrName,   
                                        bool& Rand_Flag,
                                        int& Value, 
                                        string&  Error_Text)
{
    TRACE("Input_Record::Get_Attr_As_Rand_Or_PosInt");

    list<string>  Attrib_Strings; 
    if (not Get_Attribute_Strings(AttrName, Attrib_Strings, Error_Text)) return false;
    string  String1 = Attrib_Strings.front();;
    Attrib_Strings.pop_front();
    if (not Check_No_Attrib_Strings_Left(AttrName, 
                                         Attrib_Strings, 
                                         Error_Text)) return false;
    if (String1 == "Random")
    THEN
        Rand_Flag = true;
    ELSE
        if ((not Text_To_Int(String1, Value)) or  Value < 1)
        THEN
            Error_Text = "Invalid " 
                        + AttrName 
                        + " " 
                        + String_In_Quotes(String1) 
                        + ": Expected a positive integer, or the text \"Random\"";
            return false;
        ENDIF
        Rand_Flag = false;
    ENDIF
    return true;
}
//----------------------------------------------------------------------------------------------------------------

bool Input_Record::Get_Attr_As_String(string  Attrib_Name, 
                                      string&  Value, 
                                      string&  Error_Text)
//  Get the attribute value as a single string
//  It is an error if there are multiple attribute strings

{
    TRACE("Input_Record::Get_Attrib_Value_As_Single_String");
    list<string>  Attrib_Strings;
    if (not  Get_Attribute_Strings(Attrib_Name, Attrib_Strings, Error_Text))
    THEN
        return false;
    ENDIF
    Value = Attrib_Strings.front();
    Attrib_Strings.pop_front();
    if (not Check_No_Attrib_Strings_Left(Attrib_Name, 
                                         Attrib_Strings, 
                                         Error_Text)) return false;
    return true;
}
//----------------------------------------------------------------------------------------------------------------

string Input_Record::Get_Record_Type_Text()

//  Return the Line Type Text

{
    TRACE("Input_Record::Get_Record_Type_Text");
    return Record_Type_Text;
}
//----------------------------------------------------------------------------------------------------------------

void Input_Record::Print(int  Record_Num)

//  Print the Line

{
    TRACE("Input_Record::Print");
    cout  << Record_Num <<": ";
    for (list<string>::iterator  Iter = Input_Record_For_Print.begin();
        Iter not_eq Input_Record_For_Print.end();
        Iter++)
    BEGIN
        cout << (*Iter) << "  " ;  
    END
    cout  << endl;
}
//----------------------------------------------------------------------------------------------------------------

bool  Input_Record::Read_Record(istream& Input_Stream, bool& EOF_Flag, string&  Error_Text)

//  Read the input record into a list of strings.
//  A record is terminated by a semicolon
//  Comments start with  "//" and are terminated by end of line

{
    TRACE("Input_Record::Read_Record");
    string SemiColon(";");
    string Token = ""; 
    EOF_Flag = false;
    bool End_Of_Record_Flag = false;

    while (not End_Of_Record_Flag)
    BEGIN
        if (not (Input_Stream >> Token))
        THEN
            EOF_Flag = true;
            if (not Token.empty())
            THEN
                Error_Text = Error_Text 
                            + "Incomplete record at end of file - last token is " 
                            + String_In_Quotes(Token);
                return false;
            ELSE
                return true;
            ENDIF
        ENDIF

//      Echo comment text but otherwise ignore
        if ((Token.size() > 1) 
             and (Token.substr(0,2) == "//"))
        THEN
            cout << Token;
            char Text = ' ';
            while ((Text not_eq '\n') and (not Input_Stream.eof()))
            BEGIN
                Text = Input_Stream.get(); 
                cout << Text;
            END
            Token.clear();
        ELSE
//          Detect end of record
            if (Token.substr(Token.size() -1,1) not_eq SemiColon)
            THEN
                Input_Record.push_back(Token);
            ELSE
                End_Of_Record_Flag = true;
                Token.erase(Token.size() -1,1);
                if(not Token.empty())
                THEN
                    Input_Record.push_back(Token);
                ENDIF
            ENDIF
        ENDIF
    END
    Input_Record_For_Print = Input_Record;
    return true;
}
//----------------------------------------------------------------------------------------------------------------
bool  Input_Record::Read_Record_And_Validate_Syntax(istream& Input_Stream, 
                                                    int Record_Num,
                                                    bool& EOF_Flag, 
                                                    string& Error_Text)

//  Read the next record from the file into a list of strings
//  A record is terminated by a semicolon (so that a record can span multiple lines)

//  Attribute identifiers begin with an '@' character.
//  A particular attribute identifier can occur at most once in a record.
//  An attribute can have multiple data values, and must have at least one value

{
    TRACE("Input_Record::Read_Record_And_Validate_Syntax");


    Error_Text = Get_ErrLine_Tx() + string(" Error is: ");
    if (not Read_Record(Input_Stream, EOF_Flag, Error_Text)) return false;   // Get the user input into a list of strings

//  Now populate Record_Type_Text and the Attribs_And_Data
    Attribs_And_Data.clear();
    string At("@");
//  Check the syntax of the input line and store the information
    if (not Input_Record.empty())
    THEN
        Record_Empty = false;
        bool Current_Attr_Has_No_Value = false;  // Because there is no current attribute yet!
        bool Last_Attr_Has_No_Value = false;     // Ditto
        bool Curr_Attr_Fl = false;               // Set true if there is a current attribute

    //  The first string is the Record Type    
        Record_Type_Text = Input_Record.front();
        Input_Record.pop_front();
    
    //  Get the attribute names and associated data values
        string Attrib; 
        string Data;
        while (not Input_Record.empty())
        BEGIN
            if(Input_Record.front().substr(0,1) == At)    // Check for new attribute
            THEN    
                Last_Attr_Has_No_Value = Current_Attr_Has_No_Value;
                if (Last_Attr_Has_No_Value ) break;  // Because every attribute must have at least one value
                Current_Attr_Has_No_Value = true;

                Attrib = Input_Record.front();
                Curr_Attr_Fl = true;
                Attrib = Attrib.erase(0,1);  // Remove the '@'

//              Check if the new attribute has already occurred on this line
                if (Attribs_And_Data.find(Attrib) not_eq  Attribs_And_Data.end()) 
                THEN
                    Error_Text  += string("Multiple occurrences of attribute ")
                                + String_In_Quotes(At + Attrib)
                                + string(".");;
                    return false;
                ENDIF
                Input_Record.pop_front();
            ELSE
//              Get a data value
                Data = Input_Record.front();
                Input_Record.pop_front();
                if (not Curr_Attr_Fl)  // "Value" with no attribute
                THEN
                    Error_Text =  Error_Text + string("Invalid text ")
                                + String_In_Quotes(Data)
                                + string(" is not preceeded by an attribute name.");
                    return false;
                ENDIF

                Attribs_And_Data[Attrib].push_back(Data);
                Current_Attr_Has_No_Value = false;
            ENDIF
        END

//      Check that the last attribute has at least one vaue
        if(Last_Attr_Has_No_Value or Current_Attr_Has_No_Value)  // Attribute at end of statement is 'current'
        THEN
            Error_Text  = Error_Text + string("Attribute ")
                        + String_In_Quotes(At + Attrib)
                        + string(" has no value.");
            return false;
        ENDIF
        return true;
    ELSE
        Record_Empty = true;
        return true;
    ENDIF
}
//----------------------------------------------------------------------------------------------------------------

