/****************************************************************************
*
*	  Module Filename:  Utility_Functions.cpp
*	  Module Name:      Utility_Functions
*	  Description:      General Utility Functions
*	  Author:   		Chris Gordon-Smith
*
*     Copyright (C) 2003, 2004, 2005, 2006, 2008 Chris Gordon-Smith
*     Email: See http://www.simsoup.info/
*
*     This program is free software; you can redistribute it and/or modify it
*     under the terms of the GNU General Public License Version 2, as published 
*     by the Free Software Foundation.
*
*     This program is distributed in the hope that it will be useful, but
*     WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
*     or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
*     for more details.
*
*     You should have received a copy of the GNU General Public License along
*     with this program; if not, write to the Free Software Foundation, Inc.,
*     59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*
*     See the file Licence.txt for a copy of the GNU General Public License.
*
*	  Amendment History:-
*
*		Version	   Date	   			Update Description
*		=======    ===========		==================
*
*
*
*
****************************************************************************/

//  Comment out to disallow tracing this file
//#define  ENABLE_TRACE_FILE  

#include "globldef.h"
#include "Library_Headers.h"

#include <sys/time.h>
#include <sys/resource.h>

#include  <gsl/gsl_rng.h>    // GNU Scientific library
//--------------  Utiliities  Subsystem  -------------------------------------------------------------------------
#include "Random_Sequence.h"
//----------------------------------------------------------------------------------------------------------------

namespace Utility_Functions

{   
uint Get_CPU_Time()

//  Get the time spent by the current Linux process executing user code
//  The return value is in milliseconds

{
    struct timeval tim;        
    struct rusage ru;        
    getrusage(RUSAGE_SELF, &ru);        
    tim=ru.ru_utime;        
    uint User_Time = (tim.tv_sec * 1000) + (tim.tv_usec / 1000);       
    tim=ru.ru_stime;    
    User_Time = User_Time + (tim.tv_sec * 1000) + (tim.tv_usec / 1000);       
    return User_Time;    
}
//----------------------------------------------------------------------------------------------------------------

string  Get_Equals_Tx()

{
    return string(" = ");
}
//----------------------------------------------------------------------------------------------------------------

string  Get_Error_Tx()

{
    return string("**** ERROR ****");
}
//----------------------------------------------------------------------------------------------------------------

string  Get_ErrLine_Tx()

{
    return string("****");
}
//----------------------------------------------------------------------------------------------------------------

string  Get_Warn_Tx()

{
    return string("**** WARNING ****");
}
//----------------------------------------------------------------------------------------------------------------

string  String_In_Quotes(string Tx)

//  Enclose the specified string in quotes

{
    return string("\"") + Tx + string("\"");
}
//----------------------------------------------------------------------------------------------------------------

bool Text_To_Float(string Text, double& Value)

//  Convert the specified string to a double
//  Check whether the string is valid by converting back to a string 
//  and testing for equality with the input string

{
    TRACE("Text_To_Float");

    string Std_Fmt_Tx = Text_To_Standard_Number_Format(Text);
    stringstream  StrStrm;
    StrStrm << Std_Fmt_Tx ;
    StrStrm >> Value;

    string RTString = To_String(Value);   // Get Round Trip String
    RTString = Text_To_Standard_Number_Format(RTString);

    if (RTString != Text) 
    THEN
        return false;
    ELSE
        return true;
    ENDIF
}
//----------------------------------------------------------------------------------------------------------------

bool Text_To_Int(string& Text, int& Value)

//  Check whether the Text is a valid integer in integer format 
//  and pass back value
//  Return false if text is not an integer

{ 
//  Remove any trailing blanks   

    while (Text.size() not_eq 0 and isspace(Text[Text.size()-1]))
    BEGIN
        Text.erase(Text.size()-1);
    END
  
    Value = atoi(Text.c_str());

    if (To_String(Value) != Text)  //  Check whether text entered is numeric
    THEN
        return false;
    ELSE
        return true;
    ENDIF
}
//----------------------------------------------------------------------------------------------------------

bool Text_To_PosInt(string& Text, int& Value)

//  Check whether the Text is a valid positive integer in integer format 
//  and pass back value
//  Return false if text is not an integer


{
    if (Text_To_Int(Text,Value) and Value > 0) return true;
    return false;
}
//---------------------------------------------------------------------------------------------------------
string  Text_To_Standard_Number_Format(string Text)

//  Convert a text string to standard fixed number format
//  Return an empty string if text is not a number (eg 123.6q7)

{
//  If there is a leading minus sign remove it, and then add back later.
//  This simplifies the rest of the logic

//    cout << "Text_To_Standard_Float_Format: Text = "  << Text << endl; 

    if (Text.empty()) return Text;

    char Sign = 0;
    if(Text.at(0) == '-')
    THEN
        Sign = '-';
        Text.erase(0,1);
    ENDIF

    if (Text.empty()) return Text;

//  If first character is a decimal point and it is not the only character then add a leading zero

    if (Text.at(0) == '.' and Text.size() > 1)
    THEN
        Text.insert(Text.begin(),'0');
    ENDIF

//  Remove trailing zeroes after decimal point

    if (Text.find('.') not_eq string::npos)
    THEN
        while(Text.substr(Text.size()-1) == "0")
        BEGIN
            Text.erase(Text.size()-1);
        END
    ENDIF

//  Remove decimal point if its the last character

    if (Text.substr(Text.size()-1) == ".")
    THEN
        Text.erase(Text.size()-1);
        if (Text.empty()) return Text;
    ENDIF

//  Remove leading zeroes (but leave at least one character)

    while (Text.at(0) == '0' and Text.size() != 1)
    THEN
        Text.erase(0,1);
    ENDIF
  
//  Check for any non-allowed characters

    for(uint ILoop = 0;
        ILoop < Text.size();
        ILoop++)
    BEGIN
        if((Text.substr(ILoop,1) not_eq ".")
                and
           (std::isdigit(Text.at(ILoop)) == false))
        THEN
            Text.clear();
        ENDIF
    END

//  Now add back the sign if necessary

    if(Sign)
    THEN
        Text.insert(Text.begin(),Sign);
    ENDIF

    return Text;
}
//----------------------------------------------------------------------------------------------------------------

bool    Validate_Text_As_Non_Neg_Float(string Text)

//  Check that the text specifies non negative floating point number

{
    double Value;
    if (Text_To_Float(Text, Value) and Value >= 0 )
    THEN
        return true;
    ELSE
        return false;
    ENDIF
}
//----------------------------------------------------------------------------------------------------------

bool    Validate_Text_As_Non_Neg_Int(string Text)

//  Check that the text specifies a non negative integer

{
    int Value;
    if (Text_To_Int(Text, Value) and Value >= 0)
    THEN
        return true;
    ELSE
        return false;
    ENDIF
}
//-----------------------------------------------------------------------------------------------------------------
bool    Validate_Text_As_Pos_Int(string Text)

//  Check that the text specifies a positive integer

{
    int Value;
    if (Text_To_Int(Text, Value) and Value > 0)
    THEN
        return true;
    ELSE
        return false;
    ENDIF
}
//-----------------------------------------------------------------------------------------------------------------

bool    Validate_Unique_Integers(int int1, int int2)

//  Return true if inputs are all different, otherwise false

{
    if (int1 != int2)
    THEN
        return true;
    ELSE
        return false;
    ENDIF
}
//----------------------------------------------------------------------------------------------------------
bool    Validate_Unique_Integers(int int1, int int2, int int3)

//  Return true if inputs are all different, otherwise false

{
    if ((int1 != int2) && (int2 != int3) && (int3 != int1))
    THEN
        return true;
    ELSE
        return false;
    ENDIF
} 
//----------------------------------------------------------------------------------------------------------

bool    Validate_Unique_Integers(int int1, int int2, int int3, int int4)


//  Return true if inputs are all different, otherwise false

{
    if ((int1 != int2) && (int1 != int3) && (int1 != int4)
     && (int2 != int3) && (int2 != int4)
     && (int3 != int4))
    THEN
        return true;
    ELSE
        return false;
    ENDIF
}
//----------------------------------------------------------------------------------------------------------

bool    Validate_Unique_Integers(int int1, int int2, int int3, int int4, int int5)

//  Return true if inputs are all different, otherwise false

{
    if ((int1 != int2) && (int1 != int3) && (int1 != int4) && (int1 != int5)
     && (int2 != int3) && (int2 != int4) && (int2 != int5)
     && (int3 != int4) && (int3 != int5)
     && (int4 != int5))
    THEN
        return true;
    ELSE
        return false;
    ENDIF
}
//----------------------------------------------------------------------------------------------------------


}     // end namespace
